"""Functions to generate datasets for LLM training with ballmatro hands"""
from datasets import Dataset
from itertools import combinations_with_replacement
from typing import List, Tuple, Generator, Dict, Any

from ballmatro.card import Card, SUITS, RANKS, MODIFIERS
from ballmatro.optimizer import brute_force_optimize
from ballmatro.score import ScoreInfo

def exhaustive_generator(hand_size: int) -> Generator[Tuple[List[Card], ScoreInfo]]:
    """Generator functions for a dataset with all possible hands of a given size
    and their optimal plays using brute force optimization.
    Args:
        hand_size (int): The size of the hands to generate.
    Returns:
        List[Tuple[List[Card], ScoreInfo]]: A list of tuples, each containing a hand and its optimal play in the form of a ScoreInfo object.
    """
    # Generate all possible cards
    cards = [Card(f"{rank}{suit}{modifier}") for suit in SUITS for rank in RANKS for modifier in [""] + MODIFIERS]
    
    # Generate all combinations of the given size
    for input in combinations_with_replacement(cards, hand_size):
        # Find optimal play for this input
        optimal_play = brute_force_optimize(list(input))
        yield list(input), optimal_play

def generator_to_dict(generator: Generator[Tuple[List[Card], ScoreInfo]]) -> Dict[str, List[Any]]:
    """Convert a generator of tuples to a generator of dictionaries.

    Args:
        generator (Generator[Tuple[List[Card], ScoreInfo]]): A generator that yields tuples of input cards and their corresponding ScoreInfo.

    Returns:
        Dict[str, List[Any]]: A dictionary where each key corresponds to a field in the ScoreInfo object.
    """
    # Get all the data into memory
    data = list(generator)
    # Create a dictionary with the data
    dict_data = {
        "input": [str(cards) for cards, _ in data],
        "output": [str(score.played) for _, score in data],
        "score": [score.score for _, score in data],
        "hand": [score.hand.__name__ for _, score in data],
        "chips": [score.chips for _, score in data],
        "multiplier": [score.multiplier for _, score in data],
        "remaining": [str(score.remaining) for _, score in data],
    }
    return dict_data

def to_hf_dataset(generator: Generator[Tuple[List[Card], ScoreInfo]]) -> Dataset:
    """Convert a dataset generator to a Hugging Face dataset format.
    
    Args:
        generator (Generator[Tuple[List[Card], ScoreInfo]]): A generator that yields tuples of input cards and their corresponding ScoreInfo.
    
    Returns:
        Dataset: A Hugging Face dataset containing the generated data.
    """
    # Create a Hugging Face dataset from the generator
    return Dataset.from_dict(generator_to_dict(generator))



"""Tests for the card module."""
from ballmatro.card import Card

def test_card_suit():
    card = Card("10♠")
    assert card.suit == "♠"
    card = Card("A♥")
    assert card.suit == "♥"
    card = Card("🂿")
    assert card.suit is None

def test_card_rank():
    card = Card("10♠")
    assert card.rank == "10"
    card = Card("A♥")
    assert card.rank == "A"
    card = Card("🂿Cards with rank 2 provide double chips")
    assert card.rank is None

def test_card_modifier():
    card = Card("10♠+")
    assert card.modifier == "+"
    card = Card("A♥x")
    assert card.modifier == "x"
    # card = Card("K♣*")
    # assert card.modifier == "*"
    card = Card("Q♦")
    assert card.modifier is None

def test_card_is_joker():
    card = Card("🂿Cards from the ♦ suit cannot be used in the hand")
    assert card.is_joker is True
    card = Card("10♠")
    assert card.is_joker is False

def test_card_joker_rule():
    card = Card("🂿Straights cannot be played")
    assert card.joker_rule == "Straights cannot be played"
    card = Card("10♠")
    assert card.joker_rule is None



from ballmatro.generators import exhaustive_generator, to_hf_dataset, generator_to_dict
from ballmatro.score import ScoreInfo
from ballmatro.hands import InvalidHand

def test_exhaustive_generator_size1():
    # Use a small hand size for tractable test
    results = list(exhaustive_generator(1))
    # Each result is a tuple: (hand, ScoreInfo)
    assert all(isinstance(hand, list) for hand, _ in results)
    assert all(isinstance(score_info, ScoreInfo) for _, score_info in results)
    # Check that the number of generated hands matches the expected count
    # There are 4 suits, 13 ranks, 2 modifiers, so 4*13*3 = 156 possible cards
    assert len(results) == 156
    # Check no repetitions in the generated hands
    assert len({tuple(hand) for hand, _ in results}) == len(results)
    # Check no invalid hands
    assert all(result.hand != InvalidHand for _, result in results)

def test_exhaustive_generator_size2():
    # Use a small hand size for tractable test
    results = list(exhaustive_generator(2))
    # Check that the number of generated hands matches the expected count
    # There are 4 suits, 13 ranks, 2 modifiers, so 4*13*3 = 156 possible cards
    # The number of combinations with replacement is (n + r - 1)
    # where n is the number of items to choose from and r is the number of items to choose
    # In this case, n = 156 and r = 2 so the number of combinations is (156 + 2 - 1) choose 2 = 156 * 157 / 2 = 12246
    assert len(results) == 12246
    # Check no repetitions in the generated hands
    assert len({tuple(hand) for hand, _ in results}) == len(results)
    # Check no invalid hands
    assert all(result.hand != InvalidHand for _, result in results)

def test_generator_to_dict():
    # Generate a small dataset with a small generator
    dict_generator = generator_to_dict(exhaustive_generator(1))
    # Check that the dictionary has the expected keys
    assert "input" in dict_generator
    assert "output" in dict_generator
    assert "score" in dict_generator
    assert "hand" in dict_generator
    assert "chips" in dict_generator
    assert "multiplier" in dict_generator
    assert "remaining" in dict_generator

def test_hf_dataset():
    # Generate a Hugging Face dataset with a small generator
    dataset = to_hf_dataset(exhaustive_generator(1))
    # Check that the dataset has the expected columns
    assert "input" in dataset.column_names
    assert "output" in dataset.column_names
    assert "score" in dataset.column_names
    assert "hand" in dataset.column_names
    assert "chips" in dataset.column_names
    assert "multiplier" in dataset.column_names
    assert "remaining" in dataset.column_names



import pytest

from ballmatro.card import Card
from ballmatro.optimizer import brute_force_optimize
from ballmatro.score import ScoreInfo
from ballmatro.hands import HighCard, Pair, TwoPair, ThreeOfAKind, FullHouse, Flush, Straight, FourOfAKind, StraightFlush

test_data = [
    (
        [Card('3♦')],
        ScoreInfo(played=[Card('3♦')], remaining=[], hand=HighCard, chips=8, multiplier=1)
    ),
    (
        [Card(txt='K♥x'), Card(txt='A♥x')],
        ScoreInfo(played=[Card(txt='A♥x')], remaining=[Card(txt='K♥x')], hand=HighCard, chips=16, multiplier=5)
    ),
    (
        [Card('2♥'), Card('3♦')],
        ScoreInfo(played=[Card('3♦')], remaining=[Card('2♥')], hand=HighCard, chips=8, multiplier=1)
    ),
    (
        [Card('2♥'), Card('2♥'), Card('3♦')],
        ScoreInfo(played=[Card('2♥'), Card('2♥')], remaining=[Card('3♦')], hand=Pair, chips=14, multiplier=2)
    ),
    (
        [Card('2♥'), Card('2♥'), Card('3♦'), Card('3♦'), Card('A♣')],
        ScoreInfo(played=[Card('2♥'), Card('2♥'), Card('3♦'), Card('3♦')], remaining=[Card('A♣')], hand=TwoPair, chips=30, multiplier=2)
    ),
    (
        [Card('3♥'), Card('3♦'), Card('3♠'), Card('A♣')],
        ScoreInfo(played=[Card('3♥'), Card('3♦'), Card('3♠')], remaining=[Card('A♣')], hand=ThreeOfAKind, chips=39, multiplier=3)
    ),
    (
        [Card('3♥'), Card('3♦'), Card('3♠'), Card('A♦'), Card('A♠'), Card('2♥'), Card('2♥')],
        ScoreInfo(played=[Card('3♥'), Card('3♦'), Card('3♠'), Card('A♦'), Card('A♠')], remaining=[Card('2♥'), Card('2♥')], hand=FullHouse, chips=71, multiplier=4)
    ),
    (
        [Card('2♥'), Card('2♥'), Card('3♥'), Card('5♥'), Card('8♥'), Card('J♥')],
        ScoreInfo(played=[Card('2♥'), Card('3♥'), Card('5♥'), Card('8♥'), Card('J♥')], remaining=[Card('2♥')], hand=Flush, chips=63, multiplier=4)
    ),
    (
        [Card('2♥'), Card('3♦'), Card('4♠'), Card('5♦'), Card('6♠'), Card('3♦'), Card('3♦')],
        ScoreInfo(played=[Card('2♥'), Card('3♦'), Card('4♠'), Card('5♦'), Card('6♠')], remaining=[Card('3♦'), Card('3♦')], hand=Straight, chips=50, multiplier=4)
    ),
    (
        [Card('2♥'), Card('4♥'), Card('3♥'), Card('3♦'), Card('3♠'), Card('3♣'), Card('A♥')],
        ScoreInfo(played=[Card('3♥'), Card('3♦'), Card('3♠'), Card('3♣')], remaining=[Card('2♥'), Card('4♥'), Card('A♥')], hand=FourOfAKind, chips=72, multiplier=7)
    ),
    (
        [Card('2♥'), Card('3♥'), Card('4♥'), Card('5♠'), Card('5♥'), Card('6♥'), Card('7♠')],
        ScoreInfo(played=[Card('2♥'), Card('3♥'), Card('4♥'), Card('5♥'), Card('6♥')], remaining=[Card('5♠'), Card('7♠')], hand=StraightFlush, chips=120, multiplier=8)
    ),
]

@pytest.mark.parametrize("cards, expected_score_info", test_data)
def test_brute_force_optimize(cards, expected_score_info):
    """The brute force optimizer can find the best hand for a number of cards"""
    assert brute_force_optimize(cards) == expected_score_info



"""Functions to score ballmatro hands"""
from dataclasses import dataclass
from typing import List, Tuple


from ballmatro.card import Card
from ballmatro.hands import find_hand, PokerHand, InvalidHand


CHIPS_PER_RANK = {"2": 2, "3": 3, "4": 4, "5": 5, "6": 6, "7": 7, "8": 8, "9": 9, "10": 10, "J": 10, "Q": 10, "K": 10, "A": 11}


@dataclass
class ScoreInfo:
    """Class that represents the score of a hand"""
    played: List[Card]  # Cards played in the hand
    remaining: List[Card]  # Unplayed cards
    hand: PokerHand  # Type of hand played
    chips: int = 0  # Value in chips of the hand
    multiplier: int = 0  # Multiplier for the chips value

    @property
    def score(self) -> int:
        """Return the score of the hand"""
        return self.chips * self.multiplier

    def __repr__(self):
        """Return a string representation of the score info"""
        if self.hand is None:
            return "ScoreInfo(INVALID HAND, chips=0, multiplier=0, score=0)"
        return f"ScoreInfo(played={self.played}, remaining={self.remaining}, hand={self.hand}, chips={self.chips}, multiplier={self.multiplier}, score={self.score})"

def remaining_cards(available: List[Card], played: List[Card]) -> List[Card]:
    """Returns the remaining (not played) cards after playing a hand"""
    remaining = available.copy()
    for card in played:
        # Check if the card is available
        if card not in remaining:
            raise ValueError(f"Impossible play: card {card} not in available cards")
        # Remove the card from the remaining cards
        remaining.remove(card)
    return remaining

def score_played(available: List[Card], played: List[Card]) -> ScoreInfo:
    """Given a list of played cards, find their ballmatro score
    
    A score of 0 is attained when the hand is not recognized or the list of played cards contains cards that are not available.
    """
    # Check if the played cards are available
    try:
        remaining = remaining_cards(available, played)
    except ValueError:
        # If the play is impossible, return a score of 0
        return ScoreInfo(played, available, InvalidHand)
    # Find the hand type
    hand = find_hand(played)
    if hand == InvalidHand:
        return ScoreInfo(played, remaining, InvalidHand)
    
    # Start scoring using the chips and multiplier of the hand type
    chips, multiplier = hand.chips, hand.multiplier
    # Now iterate over the cards in the order played, and score each card individually
    for card in played:
        chips, multiplier = score_card(card, chips, multiplier)

    return ScoreInfo(played, remaining, hand, chips, multiplier)

def score_card(card: Card, chips: int, multiplier: int) -> Tuple[int, int]:
    """Applies the scoring of a single card to the current chips and multiplier"""
    # Add the chips of the card rank to the current chips
    chips += CHIPS_PER_RANK.get(card.rank, 0)
    # Apply modifiers
    if card.modifier == "+":
        chips += 30
    elif card.modifier == "x":
        multiplier += 4
    return chips, multiplier
